unit FormatBIFF;

interface
uses
  Classes,SysUtils,Graphics,Dialogs;

const
  rtNone=$0000;
  rtBAM=$0001;
  rtMOS=$0002;

type
  TFileEntry=record
               FileNumber: Word; //  
               Junk_1: Word;     //
               Offset: LongInt;  //
               Size: LongInt;    //
               FType: LongInt;   //
             end;

  TFrameEntry=record
                Width: Word;
                Height: Word;
                OffsetCenterLeft: SmallInt;
                OffsetCenterTop: SmallInt;
                OffsetData: LongInt;
                Compressed: Boolean;
              end;

  TBAMImage=class
  private
  public
    BAMId: array[1..4]of char;      //
    BAMVersion: array[1..4]of char; //
    TotalFrames: Word;          // 
    TotalCycles: Byte;          // 
    TransparentColorIndex: Byte;//  
    OffsetOfEntries: LongInt;   //  
    OffsetOfPalette: LongInt;   // 
    TransColor: TColor;
    OffsetOfCycles: LongInt;
    Pal: ARRAY[0..255]OF RECORD
                           R,G,B,F: Byte;
                         END;
    Entries: array of TFrameEntry;
    Cycles: array of array of Word;
    CycleTable: array of record
                           TotalFrames: Word;
                           StartFrame: Word;
                           MaxWidth,MaxHeight: Word;
                           MinOffsetCenterLeft,MinOffsetCenterTop: SmallInt;
                         end;
    constructor Create(F: TFileStream; Offset: LongInt);
    destructor Destroy; override;
    function ExtractFrame(F: TFileStream; FBmp: TMemoryStream; Index,AnimNum: Integer; SOffset: LongInt; BackGround: TColor): Boolean;
  end;

  TFileBIFF=class
  private
  public
    F: TFileStream;
    FName: string;   // 
    FSize: LongInt; // 
    TotalFiles: LongInt; // 
    Entries: array of TFileEntry;
    constructor Create(FName: string); // 
    destructor Destroy; override;
  end;
  function DetectFileType(BIFF: TFileBIFF; Index: LongInt): Word;
  function ExtractMOS(F: TFileStream; FBmp: TMemoryStream; SOffset: LongInt): Word;
  function DetectFileById(fs: string): Word;

var
  MOSTiles: array of array of record
                                Offset: LongInt;
                                Pal: array[0..255]of record
                                                       R,G,B,F: Byte;
                                                     end;
                                Size: LongInt;
                                Width,Height: Word;
                              end;
  MOSHeader:record
              MOSId: array[1..4]of char;
              MOSVersion: array[1..4]of char;
              Width,Height: Word;
              TotalColumns: Word;
              TotalRows: Word;
              BlockSize: LongInt;
              OffsetOfPalette: LongInt;
            end;
  ID: ARRAY[1..2]OF Char;
  BMPHeader: RECORD
               FullSize: LongInt;
               _H1: LongInt;
               _C1: LongInt;
               _C2: LongInt;
               X: LongInt;
               Y: LongInt;
               _C3: Word;
               _C4: Word;
               _H2: LongInt;
               _H3: LongInt;
               DataSize1: LongInt;     //DataSize=X*Y+2*Y
               DataSize2: LongInt;     //DataSize=X*Y+2*Y
               _C5: ARRAY[1..8]OF Byte;
             END;

implementation

constructor TBAMImage.Create(F: TFileStream; Offset: LongInt);
var
  i,j: Integer;
begin
  F.Seek(Offset,soFromBeginning);
  F.Read(BAMId,4);
  F.Read(BAMVersion,4);
  F.Read(TotalFrames,2);
  F.Read(TotalCycles,1);
  F.Read(TransparentColorIndex,1);
  F.Read(OffsetOfEntries,4);
  F.Read(OffsetOfPalette,4);
  F.Seek(Offset+OffsetOfPalette,soFromBeginning);
  F.Read(Pal,1024);
  F.Seek(Offset+$14,soFromBeginning);
  F.Read(OffsetOfCycles,4);
  SetLength(Entries,TotalFrames);
  F.Seek(Offset+OffsetOfCycles,soFromBeginning);
  SetLength(Cycles,TotalCycles);
  if TotalCycles>0 then SetLength(CycleTable,TotalCycles);
  F.Seek(Offset+OffsetOfEntries,soFromBeginning);
  for i:=1 to TotalFrames do
  begin
    F.Read(Entries[i-1].Width,2);
    F.Read(Entries[i-1].Height,2);
    F.Read(Entries[i-1].OffsetCenterLeft,2);
    F.Read(Entries[i-1].OffsetCenterTop,2);
    F.Read(Entries[i-1].OffsetData,4);
    if Entries[i-1].OffsetData and $80000000=0 then Entries[i-1].Compressed:=True else Entries[i-1].Compressed:=False;
    Entries[i-1].OffsetData:=Entries[i-1].OffsetData and $7FFFFFFF;
  end;
    for i:=1 to TotalCycles do
    begin
      F.Read(CycleTable[i-1],4);
      if CycleTable[i-1].TotalFrames>0 then SetLength(Cycles[i-1],CycleTable[i-1].TotalFrames);
    end;
    F.Seek(Offset+OffsetOfCycles,soFromBeginning);
    for i:=1 to TotalCycles do
    if CycleTable[i-1].TotalFrames>0 then
    for j:=1 to CycleTable[i-1].TotalFrames do
    begin
      F.Read(Cycles[i-1][j-1],2);
      if Cycles[i-1][j-1]>TotalFrames-1 then Cycles[i-1][j-1]:=0;
    end;
    for i:=1 to TotalCycles do
    begin
      CycleTable[i-1].MaxWidth:=0;
      CycleTable[i-1].MaxHeight:=0;
      CycleTable[i-1].MinOffsetCenterLeft:=-1000;
      CycleTable[i-1].MinOffsetCenterTop:=-1000;
      for j:=1 to CycleTable[i-1].TotalFrames do
      begin
{        if Entries[Cycles[i-1][j-1]].Width>CycleTable[i-1].MaxWidth then CycleTable[i-1].MaxWidth:=Entries[Cycles[i-1][j-1]].Width;
        if Entries[Cycles[i-1][j-1]].Height>CycleTable[i-1].MaxHeight then CycleTable[i-1].MaxHeight:=Entries[Cycles[i-1][j-1]].Height;}
        if Entries[Cycles[i-1][j-1]].OffsetCenterLeft>CycleTable[i-1].MinOffsetCenterLeft then CycleTable[i-1].MinOffsetCenterLeft:=Entries[Cycles[i-1][j-1]].OffsetCenterLeft;
        if Entries[Cycles[i-1][j-1]].OffsetCenterTop>CycleTable[i-1].MinOffsetCenterTop then CycleTable[i-1].MinOffsetCenterTop:=Entries[Cycles[i-1][j-1]].OffsetCenterTop;
        // CXTP:2015.01.30
        if Entries[Cycles[i-1][j-1]].Width + (CycleTable[i-1].MinOffsetCenterLeft - Entries[Cycles[i-1][j-1]].OffsetCenterLeft) > CycleTable[i-1].MaxWidth then
          CycleTable[i-1].MaxWidth:=Entries[Cycles[i-1][j-1]].Width + (CycleTable[i-1].MinOffsetCenterLeft - Entries[Cycles[i-1][j-1]].OffsetCenterLeft);
        if Entries[Cycles[i-1][j-1]].Height + (CycleTable[i-1].MinOffsetCenterTop - Entries[Cycles[i-1][j-1]].OffsetCenterTop) > CycleTable[i-1].MaxHeight then
          CycleTable[i-1].MaxHeight:=Entries[Cycles[i-1][j-1]].Height + (CycleTable[i-1].MinOffsetCenterTop - Entries[Cycles[i-1][j-1]].OffsetCenterTop);
      end;
    end;
  TransColor:=(Pal[TransparentColorIndex].r)or(Pal[TransparentColorIndex].g shl 8)or(Pal[TransparentColorIndex].b shl 16);
end;

destructor TBAMImage.Destroy;
begin
  inherited Destroy;
end;

function DetectFileType(BIFF: TFileBIFF; Index: LongInt): Word;
begin
  DetectFileType:=rtNone;
  if (BIFF.Entries[Index].FType=$03E8)then DetectFileType:=rtBAM;
  if (BIFF.Entries[Index].FType=$03EC)then DetectFileType:=rtMOS;
end;

destructor TFileBIFF.Destroy;
begin
  if F<>nil then F.Free;
  inherited Destroy;
end;

constructor TFileBIFF.Create;
var
  i: Word;
begin
  F:=TFileStream.Create(FName,fmOpenRead);
  F.Seek($08,soFromBeginning);
  F.Read(TotalFiles,4);
  F.Seek($14,soFromBeginning);
  SetLength(Entries,TotalFiles);
  for i:=1 to TotalFiles do F.Read(Entries[i-1],16);
end;

function TBAMImage.ExtractFrame(F: TFileStream;FBmp: TMemoryStream; Index,AnimNum: Integer; SOffset: LongInt; BackGround: TColor): Boolean;
var
  Buffer: TMemoryStream;
  RowLength,i,j,LeftTopX,LeftTopY,RightBottomX,RightBottomY: LongInt;
  SType,Add: Byte;
  RType: Word;
  SmallBmp,FullBmp: TBitmap;
  TR,TG,TB: LongInt;
begin
  ExtractFrame:=True;
  if (Entries[Index].Width=0)or(Entries[Index].Height=0)then
  begin
    ExtractFrame:=False;
    Exit;
  end;
  Buffer:=TMemoryStream.Create;
  Add:=Round(Int((4*(1-(Frac(Entries[Index].Width/4))))));
  if Add=4 then Add:=0;
  Buffer.Size:=Entries[Index].Width*Entries[Index].Height+Entries[Index].Height*Add;
  F.Seek(SOffset+Entries[Index].OffsetData,soFromBeginning);
  BMPHeader.x:=Entries[Index].Width;
  BMPHeader.y:=Entries[Index].Height;
  BMPheader.DataSize1:=BMPHeader.x*BMPHeader.y;
  BMPheader.DataSize2:=BMPHeader.x*BMPHeader.y;
  BMPHeader.FullSize:=BMPHeader.DataSize1+$436;
  BMPHeader._H3:=BMPHeader.FullSize-$36;
  FBmp.Write(ID,2);
  FBmp.Write(BMPHeader,SizeOf(BMPHeader));
  FBmp.Write(Pal,SizeOf(Pal));
  {***  }
  RowLength:=0;
  for i:=1 to Entries[Index].Height do
  repeat
      F.Read(SType,1);
      if SType<>TransparentColorIndex then
      begin
        Inc(RowLength);
        Buffer.write(SType,1);
      end else
      begin
        F.Read(SType,1);
        RType:=SType+1;
        for j:=1 to RType do Buffer.write(TransparentColorIndex,1);
        Inc(RowLength,RType);
      end;
  until RowLength>=Entries[Index].Width*Entries[Index].Height;
  for i:=1 to Entries[Index].Height do
  begin
    Buffer.Seek((Entries[Index].Height-i)*Entries[Index].Width,0);
    FBmp.CopyFrom(Buffer,Entries[Index].Width+Add);
  end;
  FBmp.Seek(0,soFromBeginning);
  SmallBmp:=TBitmap.Create;
  FullBmp:=TBitmap.Create;
  SmallBmp.LoadFromStream(FBmp);
  FullBmp.Width:=CycleTable[AnimNum].MaxWidth;
  FullBmp.Height:=CycleTable[AnimNum].MaxHeight;
  LeftTopX:=CycleTable[AnimNum].MaxWidth-Entries[Index].OffsetCenterLeft-(CycleTable[AnimNum].MaxWidth-CycleTable[AnimNum].MinOffsetCenterLeft);
  LeftTopY:=CycleTable[AnimNum].MaxHeight-Entries[Index].OffsetCenterTop-(CycleTable[AnimNum].MaxHeight-CycleTable[AnimNum].MinOffsetCenterTop);
  RightBottomX:=LeftTopX+Entries[Index].Width;
  RightBottomY:=LeftTopY+Entries[Index].Height;
  FullBmp.Canvas.Brush.Color:=BackGround;
  FullBmp.Canvas.FillRect(Rect(0,0,FullBmp.Width,FullBmp.Height));
  FullBmp.Canvas.CopyRect(Rect(LeftTopX,LeftTopY,RightBottomX,RightBottomY),SmallBmp.Canvas,Rect(0,0,SmallBmp.Width,SmallBmp.Height));
  FBmp.Seek(0,soFromBeginning);
  FullBmp.SaveToStream(FBmp);
  FBmp.Seek(0,soFromBeginning);
  SmallBmp.Free;
  FullBmp.Free;
  Buffer.Free;
end;

function DetectFileById(fs: string): Word;
var
  F: TFileStream;
  TId: array[1..4]of char;
begin
  F:=TFileStream.Create(fs,fmOpenRead);
  F.Read(TId,4);
  DetectFileById:=rtNone;
  if (TId[1]='B')and(TId[2]='A')and(TId[3]='M')and(TId[4]=' ')then DetectFileById:=rtBAM;
  if (TId[1]='M')and(TId[2]='O')and(TId[3]='S')and(TId[4]=' ')then DetectFileById:=rtMOS;
  F.Free;
end;

function ExtractMOS(F: TFileStream; FBmp: TMemoryStream; SOffset: LongInt): Word;
var
  i,j,z: Integer;
  FullRows,FullColumns: Word;
  StartDataOffset,SavePos,Add: LongInt;
  TempBmp,Segment: TBitmap;
  TempStream: TMemoryStream;
begin
  TempBmp:=TBitmap.Create;
  F.Seek(SOffset,soFromBeginning);
  F.Read(MOSHeader,SizeOf(MOSHeader));
  SetLength(MOSTiles,MOSHeader.TotalRows);
  for i:=1 to MOSHeader.TotalRows do SetLength(MOSTiles[i-1],MOSHeader.TotalColumns);
  FullRows:=MOSHeader.TotalRows div MOSHeader.BlockSize;
  FullColumns:=MOSHeader.TotalColumns div MOSHeader.BlockSize;
  {***  }
  Add:=Round(Int((4*(1-(Frac(MOSHeader.Width/4))))));
  if Add=4 then Add:=0;
  for i:=1 to MOSHeader.TotalRows do
  for j:=1 to MOSHeader.TotalColumns do
  begin
    if j*MOSHeader.BlockSize<=MOSHeader.Width then MOSTiles[i-1][j-1].Width:=MOSHeader.BlockSize else MOSTiles[i-1][j-1].Width:=MOSHeader.Width-(j-1)*MOSHeader.BlockSize;
    if i*MOSHeader.BlockSize<=MOSHeader.Height then MOSTiles[i-1][j-1].Height:=MOSHeader.BlockSize else MOSTiles[i-1][j-1].Height:=MOSHeader.Height-(i-1)*MOSHeader.BlockSize;
  end;
  {***  }
  F.Seek(SOffset+MOSHeader.OffsetOfPalette,soFromBeginning);
  for i:=1 to MOSHeader.TotalRows do
  for j:=1 to MOSHeader.TotalColumns do F.Read(MOSTiles[i-1][j-1].Pal,1024);
  {***   }
  for i:=1 to MOSHeader.TotalRows do
  for j:=1 to MOSHeader.TotalColumns do F.Read(MOSTiles[i-1][j-1].Offset,4);
  StartDataOffset:=F.Position;
  TempBmp.Width:=MOSHeader.Width;
  TempBmp.Height:=MOSHeader.Height;
  {***    [1,1]}
  Segment:=TBitmap.Create;
  for i:=1 to MOSHeader.TotalRows do
  for j:=1 to MOSHeader.TotalColumns do
  begin
    TempStream:=TMemoryStream.Create;
    BMPHeader.x:=MOSTiles[i-1][j-1].Width;
    BMPHeader.y:=MOSTiles[i-1][j-1].Height;
    BMPheader.DataSize1:=BMPHeader.x*BMPHeader.y;
    BMPheader.DataSize2:=BMPHeader.x*BMPHeader.y;
    BMPHeader.FullSize:=BMPHeader.DataSize1+$436;
    BMPHeader._H3:=BMPHeader.FullSize-$36;
    TempStream.Write(ID,2);
    TempStream.Write(BMPHeader,SizeOf(BMPHeader));
    TempStream.Write(MOSTiles[i-1][j-1].Pal,1024);
    F.Seek(SOffset+StartDataOffset+MOSTiles[i-1][j-1].Offset,soFromBeginning);
    SavePos:=SOffset+$18+1024*MOSHeader.TotalColumns*MOSHeader.TotalRows+4*MOSHeader.TotalColumns*MOSHeader.TotalRows+MOSTiles[i-1][j-1].Offset;
    for z:=MOSTiles[i-1][j-1].Height downto 1 do
    begin
      F.Seek(SavePos+(z-1)*MOSTiles[i-1][j-1].Width,soFromBeginning);
      TempStream.CopyFrom(F,MOSTiles[i-1][j-1].Width);
      if (j=MOSHeader.TotalColumns)and(Add>0)then TempStream.Write(Add,Add);
    end;
    TempStream.Seek($00,soFromBeginning);
    Segment.LoadFromStream(TempStream);
    TempBmp.Canvas.CopyRect(Rect((j-1)*MOSHeader.BlockSize,(i-1)*MOSHeader.BlockSize,(j-1)*MOSHeader.BlockSize+MOSTiles[i-1][j-1].Width,(i-1)*MOSHeader.BlockSize+MOSTiles[i-1][j-1].Height),Segment.Canvas,Rect(0,0,Segment.Width,Segment.Height));
    TempStream.Free;
  end;
  Segment.Free;
  TempBmp.SaveToStream(FBmp);
  FBmp.Seek($00,soFromBeginning);
  {***    [1,1]}
  TempBmp.Free;
end;

begin
  ID[1]:='B';
  ID[2]:='M';
  BMPHeader._C1:=$00000436;
  BMPHeader._C2:=$00000028;
  BMPHeader._C3:=$0001;
  BMPHeader._C4:=$0008;
  BMPHeader._H2:=$00;
  BMPHeader._H3:=$00;
end.
